@@ -83,9 +83,8 @@
 #define BUF_SIZE 2048
 #endif
 
-int verbose        = 0;
-int reuse_port     = 0;
-int keep_resolving = 1;
+int verbose    = 0;
+int reuse_port = 0;
 
 #ifdef __ANDROID__
 int vpn        = 0;
@@ -277,6 +276,60 @@ delayed_connect_cb(EV_P_ ev_timer *watcher, int revents)
     server_recv_cb(EV_A_ & server->recv_ctx->io, revents);
 }
 
+static int
+server_handshake_reply(EV_P_ ev_io *w, int udp_assc, struct socks5_response *response)
+{
+    server_ctx_t *server_recv_ctx = (server_ctx_t *)w;
+    server_t *server              = server_recv_ctx->server;
+    remote_t *remote              = server->remote;
+    if (server->stage != STAGE_HANDSHAKE)
+        return 0;
+
+    struct sockaddr_in sock_addr;
+    if (udp_assc) {
+        socklen_t addr_len = sizeof(sock_addr);
+        if (getsockname(server->fd, (struct sockaddr *)&sock_addr, &addr_len) < 0) {
+            LOGE("getsockname: %s", strerror(errno));
+            response->rep = SOCKS5_REP_CONN_REFUSED;
+            send(server->fd, (char *)response, sizeof(struct socks5_response), 0);
+            close_and_free_remote(EV_A_ remote);
+            close_and_free_server(EV_A_ server);
+            return -1;
+        }
+    } else
+        memset(&sock_addr, 0, sizeof(sock_addr));
+
+    buffer_t resp_to_send;
+    buffer_t *resp_buf = &resp_to_send;
+    balloc(resp_buf, BUF_SIZE);
+
+    memcpy(resp_buf->data, response, sizeof(struct socks5_response));
+    memcpy(resp_buf->data + sizeof(struct socks5_response),
+           &sock_addr.sin_addr, sizeof(sock_addr.sin_addr));
+    memcpy(resp_buf->data + sizeof(struct socks5_response) +
+           sizeof(sock_addr.sin_addr),
+           &sock_addr.sin_port, sizeof(sock_addr.sin_port));
+
+    int reply_size = sizeof(struct socks5_response) +
+                     sizeof(sock_addr.sin_addr) + sizeof(sock_addr.sin_port);
+
+    int s = send(server->fd, resp_buf->data, reply_size, 0);
+
+    bfree(resp_buf);
+
+    if (s < reply_size) {
+        LOGE("failed to send fake reply");
+        close_and_free_remote(EV_A_ remote);
+        close_and_free_server(EV_A_ server);
+        return -1;
+    }
+    if (udp_assc) {
+        // Wait until client closes the connection
+        return -1;
+    }
+    return 0;
+}
+
 static int
 server_handshake(EV_P_ ev_io *w, buffer_t *buf)
 {
@@ -286,30 +339,25 @@ server_handshake(EV_P_ ev_io *w, buffer_t *buf)
 
     struct socks5_request *request = (struct socks5_request *)buf->data;
     size_t request_len             = sizeof(struct socks5_request);
-    struct sockaddr_in sock_addr;
-    memset(&sock_addr, 0, sizeof(sock_addr));
 
     if (buf->len < request_len) {
         return -1;
     }
 
-    int udp_assc = 0;
+    struct socks5_response response;
+    response.ver  = SVERSION;
+    response.rep  = SOCKS5_REP_SUCCEEDED;
+    response.rsv  = 0;
+    response.atyp = SOCKS5_ATYP_IPV4;
 
-    if (request->cmd == 3) {
-        udp_assc = 1;
-        socklen_t addr_len = sizeof(sock_addr);
-        getsockname(udp_fd, (struct sockaddr *)&sock_addr,
-                    &addr_len);
+    if (request->cmd == SOCKS5_CMD_UDP_ASSOCIATE) {
         if (verbose) {
             LOGI("udp assc request accepted");
         }
-    } else if (request->cmd != 1) {
+        return server_handshake_reply(EV_A_ w, 1, &response);
+    } else if (request->cmd != SOCKS5_CMD_CONNECT) {
         LOGE("unsupported cmd: %d", request->cmd);
-        struct socks5_response response;
-        response.ver  = SVERSION;
-        response.rep  = CMD_NOT_SUPPORTED;
-        response.rsv  = 0;
-        response.atyp = 1;
+        response.rep = SOCKS5_REP_CMD_NOT_SUPPORTED;
         char *send_buf = (char *)&response;
         send(server->fd, send_buf, 4, 0);
         close_and_free_remote(EV_A_ remote);
@@ -317,47 +365,7 @@ server_handshake(EV_P_ ev_io *w, buffer_t *buf)
         return -1;
     }
 
-    // Fake reply
-    if (server->stage == STAGE_HANDSHAKE) {
-        struct socks5_response response;
-        response.ver  = SVERSION;
-        response.rep  = 0;
-        response.rsv  = 0;
-        response.atyp = 1;
-
-        buffer_t resp_to_send;
-        buffer_t *resp_buf = &resp_to_send;
-        balloc(resp_buf, BUF_SIZE);
-
-        memcpy(resp_buf->data, &response, sizeof(struct socks5_response));
-        memcpy(resp_buf->data + sizeof(struct socks5_response),
-               &sock_addr.sin_addr, sizeof(sock_addr.sin_addr));
-        memcpy(resp_buf->data + sizeof(struct socks5_response) +
-               sizeof(sock_addr.sin_addr),
-               &sock_addr.sin_port, sizeof(sock_addr.sin_port));
-
-        int reply_size = sizeof(struct socks5_response) +
-                         sizeof(sock_addr.sin_addr) + sizeof(sock_addr.sin_port);
-
-        int s = send(server->fd, resp_buf->data, reply_size, 0);
-
-        bfree(resp_buf);
-
-        if (s < reply_size) {
-            LOGE("failed to send fake reply");
-            close_and_free_remote(EV_A_ remote);
-            close_and_free_server(EV_A_ server);
-            return -1;
-        }
-        if (udp_assc) {
-            // Wait until client closes the connection
-            return -1;
-        }
-
-        server->stage = STAGE_PARSE;
-    }
-
-    char host[257], ip[INET6_ADDRSTRLEN], port[16];
+    char host[MAX_HOSTNAME_LEN+1], ip[INET6_ADDRSTRLEN], port[16];
 
     buffer_t *abuf = server->abuf;
     abuf->idx = 0;
@@ -367,8 +375,7 @@ server_handshake(EV_P_ ev_io *w, buffer_t *buf)
     int atyp = request->atyp;
 
     // get remote addr and port
-    if (atyp == 1) {
-        // IP V4
+    if (atyp == SOCKS5_ATYP_IPV4) {
         size_t in_addr_len = sizeof(struct in_addr);
         if (buf->len < request_len + in_addr_len + 2) {
             return -1;
@@ -378,12 +385,14 @@ server_handshake(EV_P_ ev_io *w, buffer_t *buf)
 
         if (acl || verbose) {
             uint16_t p = ntohs(*(uint16_t *)(buf->data + request_len + in_addr_len));
-            inet_ntop(AF_INET, (const void *)(buf->data + request_len),
-                      ip, INET_ADDRSTRLEN);
+            if (!inet_ntop(AF_INET, (const void *)(buf->data + request_len),
+                           ip, INET_ADDRSTRLEN)) {
+                LOGI("inet_ntop(AF_INET): %s", strerror(errno));
+                ip[0] = '\0';
+            }
             sprintf(port, "%d", p);
         }
-    } else if (atyp == 3) {
-        // Domain name
+    } else if (atyp == SOCKS5_ATYP_DOMAIN) {
         uint8_t name_len = *(uint8_t *)(buf->data + request_len);
         if (buf->len < request_len + 1 + name_len + 2) {
             return -1;
@@ -399,8 +408,7 @@ server_handshake(EV_P_ ev_io *w, buffer_t *buf)
             host[name_len] = '\0';
             sprintf(port, "%d", p);
         }
-    } else if (atyp == 4) {
-        // IP V6
+    } else if (atyp == SOCKS5_ATYP_IPV6) {
         size_t in6_addr_len = sizeof(struct in6_addr);
         if (buf->len < request_len + in6_addr_len + 2) {
             return -1;
@@ -410,12 +418,18 @@ server_handshake(EV_P_ ev_io *w, buffer_t *buf)
 
         if (acl || verbose) {
             uint16_t p = ntohs(*(uint16_t *)(buf->data + request_len + in6_addr_len));
-            inet_ntop(AF_INET6, (const void *)(buf->data + request_len),
-                      ip, INET6_ADDRSTRLEN);
+            if (!inet_ntop(AF_INET6, (const void *)(buf->data + request_len),
+                           ip, INET6_ADDRSTRLEN)) {
+                LOGI("inet_ntop(AF_INET6): %s", strerror(errno));
+                ip[0] = '\0';
+            }
             sprintf(port, "%d", p);
         }
     } else {
         LOGE("unsupported addrtype: %d", request->atyp);
+        response.rep = SOCKS5_REP_ADDRTYPE_NOT_SUPPORTED;
+        char *send_buf = (char *)&response;
+        send(server->fd, send_buf, 4, 0);
         close_and_free_remote(EV_A_ remote);
         close_and_free_server(EV_A_ server);
         return -1;
@@ -423,32 +437,37 @@ server_handshake(EV_P_ ev_io *w, buffer_t *buf)
 
     size_t abuf_len  = abuf->len;
     int sni_detected = 0;
-    int ret          = 0;
+    int hostname_len = 0;
 
     char *hostname;
     uint16_t dst_port = ntohs(*(uint16_t *)(abuf->data + abuf->len - 2));
 
-    if (atyp == 1 || atyp == 4) {
+    if (atyp == SOCKS5_ATYP_IPV4 || atyp == SOCKS5_ATYP_IPV6) {
         if (dst_port == http_protocol->default_port)
-            ret = http_protocol->parse_packet(buf->data + 3 + abuf->len,
+            hostname_len = http_protocol->parse_packet(buf->data + 3 + abuf->len,
                                               buf->len - 3 - abuf->len, &hostname);
         else if (dst_port == tls_protocol->default_port)
-            ret = tls_protocol->parse_packet(buf->data + 3 + abuf->len,
+            hostname_len = tls_protocol->parse_packet(buf->data + 3 + abuf->len,
                                              buf->len - 3 - abuf->len, &hostname);
-        if (ret == -1 && buf->len < BUF_SIZE && server->stage != STAGE_SNI) {
+        if (hostname_len == -1 && buf->len < BUF_SIZE && server->stage != STAGE_SNI) {
+            if (server_handshake_reply(EV_A_ w, 0, &response) < 0)
+                return -1;
             server->stage = STAGE_SNI;
             ev_timer_start(EV_A_ & server->delayed_connect_watcher);
             return -1;
-        } else if (ret > 0) {
+        } else if (hostname_len > 0) {
             sni_detected = 1;
             if (acl || verbose) {
-                memcpy(host, hostname, ret);
-                host[ret] = '\0';
+                hostname_len = hostname_len > MAX_HOSTNAME_LEN ? MAX_HOSTNAME_LEN : hostname_len;
+                memcpy(host, hostname, hostname_len);
+                host[hostname_len] = '\0';
             }
             ss_free(hostname);
         }
     }
 
+    if (server_handshake_reply(EV_A_ w, 0, &response) < 0)
+        return -1;
     server->stage = STAGE_STREAM;
 
     buf->len -= (3 + abuf_len);
@@ -457,11 +476,11 @@ server_handshake(EV_P_ ev_io *w, buffer_t *buf)
     }
 
     if (verbose) {
-        if (sni_detected || atyp == 3)
+        if (sni_detected || atyp == SOCKS5_ATYP_DOMAIN)
             LOGI("connect to %s:%s", host, port);
-        else if (atyp == 1)
+        else if (atyp == SOCKS5_ATYP_IPV4)
             LOGI("connect to %s:%s", ip, port);
-        else if (atyp == 4)
+        else if (atyp == SOCKS5_ATYP_IPV6)
             LOGI("connect to [%s]:%s", ip, port);
     }
 
@@ -471,15 +490,13 @@ server_handshake(EV_P_ ev_io *w, buffer_t *buf)
 #endif
         ) {
         int bypass   = 0;
-#ifndef __ANDROID__
         int resolved = 0;
-#endif
         struct sockaddr_storage storage;
         memset(&storage, 0, sizeof(struct sockaddr_storage));
         int err;
 
         int host_match = 0;
-        if (sni_detected || atyp == 3)
+        if (sni_detected || atyp == SOCKS5_ATYP_DOMAIN)
             host_match = acl_match_host(host);
 
         if (host_match > 0)
@@ -487,31 +504,37 @@ server_handshake(EV_P_ ev_io *w, buffer_t *buf)
         else if (host_match < 0)
             bypass = 0;                             // proxy hostnames in white list
         else {
-#ifndef __ANDROID__
-            if (atyp == 3) {                        // resolve domain so we can bypass domain with geoip
-                err = get_sockaddr(host, port, &storage, 0, ipv6first);
-                if (err != -1) {
-                    resolved = 1;
-                    switch (((struct sockaddr *)&storage)->sa_family) {
-                    case AF_INET:
-                    {
-                        struct sockaddr_in *addr_in = (struct sockaddr_in *)&storage;
-                        inet_ntop(AF_INET, &(addr_in->sin_addr), ip, INET_ADDRSTRLEN);
-                        break;
-                    }
-                    case AF_INET6:
-                    {
-                        struct sockaddr_in6 *addr_in6 = (struct sockaddr_in6 *)&storage;
-                        inet_ntop(AF_INET6, &(addr_in6->sin6_addr), ip, INET6_ADDRSTRLEN);
-                        break;
-                    }
-                    default:
-                        break;
-                    }
+            if (atyp == SOCKS5_ATYP_DOMAIN
+#ifdef __ANDROID__
+                && !vpn
+#endif
+                ) {           // resolve domain so we can bypass domain with geoip
+                if (get_sockaddr(host, port, &storage, 0, ipv6first))
+                    goto not_bypass;
+                resolved = 1;
+                switch (((struct sockaddr *)&storage)->sa_family) {
+                case AF_INET:
+                {
+                    struct sockaddr_in *addr_in = (struct sockaddr_in *)&storage;
+                    if (!inet_ntop(AF_INET, &(addr_in->sin_addr), ip, INET_ADDRSTRLEN))
+                        goto not_bypass;
+                    break;
+                }
+                case AF_INET6:
+                {
+                    struct sockaddr_in6 *addr_in6 = (struct sockaddr_in6 *)&storage;
+                    if (!inet_ntop(AF_INET6, &(addr_in6->sin6_addr), ip, INET6_ADDRSTRLEN))
+                        goto not_bypass;
+                    break;
+                }
+                default:
+                    goto not_bypass;
                 }
             }
-#endif
-            int ip_match = acl_match_host(ip);
+
+            int ip_match = (resolved || atyp == SOCKS5_ATYP_IPV4
+                    || atyp == SOCKS5_ATYP_IPV6) ? acl_match_host(ip) : 0;
+
             switch (get_acl_mode()) {
             case BLACK_LIST:
                 if (ip_match > 0)
@@ -527,19 +550,22 @@ server_handshake(EV_P_ ev_io *w, buffer_t *buf)
 
         if (bypass) {
             if (verbose) {
-                if (sni_detected || atyp == 3)
+                if (sni_detected || atyp == SOCKS5_ATYP_DOMAIN)
                     LOGI("bypass %s:%s", host, port);
                 else if (atyp == 1)
                     LOGI("bypass %s:%s", ip, port);
                 else if (atyp == 4)
                     LOGI("bypass [%s]:%s", ip, port);
             }
-#ifndef __ANDROID__
-            if (atyp == 3 && resolved != 1)
+            if (atyp == SOCKS5_ATYP_DOMAIN && !resolved)
+#ifdef __ANDROID__
+                if (vpn)
+                    goto not_bypass;
+                else
+#endif
                 err = get_sockaddr(host, port, &storage, 0, ipv6first);
             else
-#endif
-            err = get_sockaddr(ip, port, &storage, 0, ipv6first);
+                err = get_sockaddr(ip, port, &storage, 0, ipv6first);
             if (err != -1) {
                 remote = create_remote(server->listener, (struct sockaddr *)&storage);
                 if (remote != NULL)
@@ -548,6 +574,7 @@ server_handshake(EV_P_ ev_io *w, buffer_t *buf)
         }
     }
 
+not_bypass:
     // Not bypass
     if (remote == NULL) {
         remote = create_remote(server->listener, NULL);
@@ -560,9 +587,9 @@ server_handshake(EV_P_ ev_io *w, buffer_t *buf)
             // Reconstruct address buffer
             abuf->len               = 0;
             abuf->data[abuf->len++] = 3;
-            abuf->data[abuf->len++] = ret;
-            memcpy(abuf->data + abuf->len, host, ret);
-            abuf->len += ret;
+            abuf->data[abuf->len++] = hostname_len;
+            memcpy(abuf->data + abuf->len, host, hostname_len);
+            abuf->len += hostname_len;
             dst_port   = htons(dst_port);
             memcpy(abuf->data + abuf->len, &dst_port, 2);
             abuf->len += 2;
@@ -615,6 +642,8 @@ server_stream(EV_P_ ev_io *w, buffer_t *buf)
         return;
     }
 
+    ev_timer_again(EV_A_ & remote->recv_ctx->watcher);
+
     // insert shadowsocks header
     if (!remote->direct) {
 #ifdef __ANDROID__
@@ -897,7 +926,6 @@ server_recv_cb(EV_P_ ev_io *w, int revents)
             buf->len = 0;
             return;
         } else if (server->stage == STAGE_HANDSHAKE ||
-                   server->stage == STAGE_PARSE ||
                    server->stage == STAGE_SNI) {
             int ret = server_handshake(EV_A_ w, buf);
             if (ret)
@@ -1371,7 +1399,6 @@ signal_cb(EV_P_ ev_signal *w, int revents)
             ev_io_stop(EV_DEFAULT, &plugin_watcher.io);
 #endif
 #endif
-            keep_resolving = 0;
             ev_unloop(EV_A_ EVUNLOOP_ALL);
         }
     }
@@ -1393,7 +1420,6 @@ plugin_watcher_cb(EV_P_ ev_io *w, int revents)
     ev_signal_stop(EV_DEFAULT, &sigint_watcher);
     ev_signal_stop(EV_DEFAULT, &sigterm_watcher);
     ev_io_stop(EV_DEFAULT, &plugin_watcher.io);
-    keep_resolving = 0;
     ev_unloop(EV_A_ EVUNLOOP_ALL);
 }
 
@@ -1697,7 +1723,7 @@ main(int argc, char **argv)
     }
 
     if (method == NULL) {
-        method = "rc4-md5";
+        method = "chacha20-ietf-poly1305";
     }
 
     if (timeout == NULL) {
@@ -1850,7 +1876,7 @@ main(int argc, char **argv)
     ev_signal_start(EV_DEFAULT, &sigchld_watcher);
 #endif
 
-    if (strcmp(local_addr, ":") > 0)
+    if (ss_is_ipv6addr(local_addr))
         LOGI("listening at [%s]:%s", local_addr, local_port);
     else
         LOGI("listening at %s:%s", local_addr, local_port);
@@ -2032,7 +2058,7 @@ _start_ss_local_server(profile_t profile, ss_local_callback callback, void *udat
     listen_ctx.iface          = NULL;
     listen_ctx.mptcp          = mptcp;
 
-    if (strcmp(local_addr, ":") > 0)
+    if (ss_is_ipv6addr(local_addr))
         LOGI("listening at [%s]:%s", local_addr, local_port_str);
     else
         LOGI("listening at %s:%s", local_addr, local_port_str);
