@@ -232,6 +232,7 @@ stat_update_cb(EV_P_ ev_timer *watcher, int revents)
 
     close(sfd);
 }
+
 #endif
 
 static void
@@ -345,6 +346,7 @@ setnonblocking(int fd)
     }
     return fcntl(fd, F_SETFL, flags | O_NONBLOCK);
 }
+
 #endif
 
 int
@@ -362,23 +364,15 @@ create_and_bind(const char *host, const char *port, int mptcp)
 
     result = NULL;
 
-    for (int i = 1; i < 8; i++) {
-        s = getaddrinfo(host, port, &hints, &result);
-        if (s == 0) {
-            break;
-        } else {
-            sleep(pow(2, i));
-            LOGE("failed to resolve server name, wait %.0f seconds", pow(2, i));
-        }
-    }
+    s = getaddrinfo(host, port, &hints, &result);
 
     if (s != 0) {
-        LOGE("getaddrinfo: %s", gai_strerror(s));
+        LOGE("failed to resolve server name %s", host);
         return -1;
     }
 
     if (result == NULL) {
-        LOGE("Could not bind");
+        LOGE("Cannot bind");
         return -1;
     }
 
@@ -528,9 +522,9 @@ connect_to_remote(EV_P_ struct addrinfo *res,
 #if defined(MSG_FASTOPEN) && !defined(TCP_FASTOPEN_CONNECT)
         int s = -1;
         s = sendto(sockfd, server->buf->data, server->buf->len,
-                MSG_FASTOPEN, res->ai_addr, res->ai_addrlen);
+                   MSG_FASTOPEN, res->ai_addr, res->ai_addrlen);
 #elif defined(TCP_FASTOPEN_WINSOCK)
-        DWORD s = -1;
+        DWORD s   = -1;
         DWORD err = 0;
         do {
             int optval = 1;
@@ -560,14 +554,14 @@ connect_to_remote(EV_P_ struct addrinfo *res,
                           &s, &remote->olap)) {
                 remote->connect_ex_done = 1;
                 break;
-            };
+            }
             // XXX: ConnectEx pending, check later in remote_send
             if (WSAGetLastError() == ERROR_IO_PENDING) {
                 err = CONNECT_IN_PROGRESS;
                 break;
             }
             ERROR("ConnectEx");
-        } while(0);
+        } while (0);
         // Set error number
         if (err) {
             SetLastError(err);
@@ -576,8 +570,8 @@ connect_to_remote(EV_P_ struct addrinfo *res,
         int s = -1;
 #if defined(TCP_FASTOPEN_CONNECT)
         int optval = 1;
-        if(setsockopt(sockfd, IPPROTO_TCP, TCP_FASTOPEN_CONNECT,
-                    (void *)&optval, sizeof(optval)) < 0)
+        if (setsockopt(sockfd, IPPROTO_TCP, TCP_FASTOPEN_CONNECT,
+                       (void *)&optval, sizeof(optval)) < 0)
             FATAL("failed to set TCP_FASTOPEN_CONNECT");
         s = connect(sockfd, res->ai_addr, res->ai_addrlen);
 #elif defined(CONNECT_DATA_IDEMPOTENT)
@@ -588,7 +582,7 @@ connect_to_remote(EV_P_ struct addrinfo *res,
         endpoints.sae_dstaddrlen = res->ai_addrlen;
 
         s = connectx(sockfd, &endpoints, SAE_ASSOCID_ANY, CONNECT_DATA_IDEMPOTENT,
-                         NULL, 0, NULL, NULL);
+                     NULL, 0, NULL, NULL);
 #else
         FATAL("fast open is not enabled in this build");
 #endif
@@ -933,9 +927,9 @@ server_recv_cb(EV_P_ ev_io *w, int revents)
 
         if (verbose) {
             if ((atyp & ADDRTYPE_MASK) == 4)
-                LOGI("connect to [%s]:%d", host, ntohs(port));
+                LOGI("[%s] connect to [%s]:%d", remote_port, host, ntohs(port));
             else
-                LOGI("connect to %s:%d", host, ntohs(port));
+                LOGI("[%s] connect to %s:%d", remote_port, host, ntohs(port));
         }
 
         if (!need_query) {
@@ -1236,8 +1230,8 @@ remote_send_cb(EV_P_ ev_io *w, int revents)
                 // Non-blocking way to fetch ConnectEx result
                 if (WSAGetOverlappedResult(remote->fd, &remote->olap,
                                            &numBytes, FALSE, &flags)) {
-                    remote->buf->len -= numBytes;
-                    remote->buf->idx  = numBytes;
+                    remote->buf->len       -= numBytes;
+                    remote->buf->idx        = numBytes;
                     remote->connect_ex_done = 1;
                 } else if (WSAGetLastError() == WSA_IO_INCOMPLETE) {
                     // XXX: ConnectEx still not connected, wait for next time
@@ -1248,7 +1242,7 @@ remote_send_cb(EV_P_ ev_io *w, int revents)
                     close_and_free_remote(EV_A_ remote);
                     close_and_free_server(EV_A_ server);
                     return;
-                };
+                }
             }
 
             // Make getpeername work
@@ -1506,8 +1500,7 @@ signal_cb(EV_P_ ev_signal *w, int revents)
             if (!is_plugin_running()) {
                 LOGE("plugin service exit unexpectedly");
                 ret_val = -1;
-            }
-            else
+            } else
                 return;
 #endif
         case SIGINT:
@@ -1542,6 +1535,7 @@ plugin_watcher_cb(EV_P_ ev_io *w, int revents)
     ev_io_stop(EV_DEFAULT, &plugin_watcher.io);
     ev_unloop(EV_A_ EVUNLOOP_ALL);
 }
+
 #endif
 
 static void
@@ -1860,7 +1854,7 @@ main(int argc, char **argv)
     }
 
     if (method == NULL) {
-        method = "rc4-md5";
+        method = "chacha20-ietf-poly1305";
     }
 
     if (timeout == NULL) {
@@ -1954,9 +1948,9 @@ main(int argc, char **argv)
             do {
                 struct sockaddr_in addr;
                 memset(&addr, 0, sizeof(addr));
-                addr.sin_family = AF_INET;
+                addr.sin_family      = AF_INET;
                 addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
-                addr.sin_port = htons(plugin_watcher.port);
+                addr.sin_port        = htons(plugin_watcher.port);
                 if (bind(fd, (struct sockaddr *)&addr, sizeof(addr))) {
                     LOGE("failed to bind plugin control port");
                     break;
@@ -2008,6 +2002,7 @@ main(int argc, char **argv)
 
     // bind to each interface
     if (mode != UDP_ONLY) {
+        int num_listen_ctx = 0;
         for (int i = 0; i < server_num; i++) {
             const char *host = server_host[i];
 
@@ -2015,7 +2010,7 @@ main(int argc, char **argv)
                 host = "127.0.0.1";
             }
 
-            if (host && strcmp(host, ":") > 0)
+            if (host && ss_is_ipv6addr(host))
                 LOGI("tcp server listening at [%s]:%s", host, server_port);
             else
                 LOGI("tcp server listening at %s:%s", host ? host : "0.0.0.0", server_port);
@@ -2024,10 +2019,11 @@ main(int argc, char **argv)
             int listenfd;
             listenfd = create_and_bind(host, server_port, mptcp);
             if (listenfd == -1) {
-                FATAL("bind() error");
+                continue;
             }
             if (listen(listenfd, SSMAXCONN) == -1) {
-                FATAL("listen() error");
+                ERROR("listen()");
+                continue;
             }
             setfastopen(listenfd);
             setnonblocking(listenfd);
@@ -2042,24 +2038,37 @@ main(int argc, char **argv)
             ev_io_init(&listen_ctx->io, accept_cb, listenfd, EV_READ);
             ev_io_start(loop, &listen_ctx->io);
 
+            num_listen_ctx++;
+
             if (plugin != NULL)
                 break;
         }
+
+        if (num_listen_ctx == 0) {
+            FATAL("failed to listen on any address");
+        }
     }
 
     if (mode != TCP_ONLY) {
+        int num_listen_ctx = 0;
         for (int i = 0; i < server_num; i++) {
             const char *host = server_host[i];
             const char *port = server_port;
             if (plugin != NULL) {
                 port = plugin_port;
             }
-            if (host && strcmp(host, ":") > 0)
+            if (host && ss_is_ipv6addr(host))
                 LOGI("udp server listening at [%s]:%s", host, port);
             else
                 LOGI("udp server listening at %s:%s", host ? host : "0.0.0.0", port);
             // Setup UDP
-            init_udprelay(host, port, mtu, crypto, atoi(timeout), iface);
+            int err = init_udprelay(host, port, mtu, crypto, atoi(timeout), iface);
+            if (err == -1) continue;
+            num_listen_ctx++;
+        }
+
+        if (num_listen_ctx == 0) {
+            FATAL("failed to listen on any address");
         }
     }
 
